import sys, random
import os
from ImageD11.grid_index_parallel import grid_index_parallel

if __name__=="__main__":
    # You need this idiom to use multiprocessing on windows (script is imported again)
    
    symmetry = sys.argv[9]
    ring1 = list(map(int, sys.argv[10].split(',')))
    ring2 = list(map(int, sys.argv[11].split(',')))
    
    gridpars = {
        'DSTOL' : 0.004,
        'OMEGAFLOAT' : 0.13,
        'COSTOL' : 0.002,
        'NPKS' : int(sys.argv[4]),
        'TOLSEQ' : [float(sys.argv[6]), float(sys.argv[7])],
        'SYMMETRY' : symmetry,
        'RING1'  : ring1,
        'RING2' : ring2,
        'NUL' : True,
        'FITPOS' : True,
        'tolangle' : 1.0,
        'toldist' : int(sys.argv[8]),
        'NPROC' : None,
        'NTHREAD' : 1,
    }

    try:
        gridpars['NUNIQ'] = int(sys.argv[5])
    except:
        pass
    # grid to search
    r = 800    # radius of cylinder
    r2 = 500   # Beam height
    translations = [(t_x, t_y, t_z)
        for t_x in range(-r, r+1, 200)
        for t_y in range(-r, r+1, 200)
        for t_z in range(-r2, r2+1, 200) ]
    # Cylinder:
    translations = [( x,y,z) for (x,y,z) in translations if (x*x+y*y)< r*r ]
    #
    random.seed(42) # reproducible
    random.shuffle(translations)

    fltfile = sys.argv[1]
    parfile = sys.argv[2]
    tmp = sys.argv[3]
    
    grid_index_parallel( fltfile, parfile, tmp, gridpars, translations)


import sys
import numpy as np
import time
import os
import functools
import numpy as np
import hdf5plugin
import h5py
import numba
import fabio

zcenter = 1056
bg_scale = 0

removalmask = np.ones((2048,2048))

for ii in range(2048):
    for jj in range(2048):
        disttmp = np.sqrt((ii-zcenter)**2+(jj-1005)**2) # works for diff1-scan1-2. 1054 and 1008 are the centers for this
        if (disttmp > 480) and (disttmp < 730) or (disttmp>930):     #(disttmp > 380) and (disttmp < 730) or (disttmp>930):
            removalmask[ii,jj] = 1.0    # 0 if removal is true

# Code to clean the 2D image and reduce it to a sparse array:
# things we might edit
class Options:
    HOWMANY = int(2048*2048*0.25)  # max pixels per frame to keep : 10%
    # additional thresholds to test if you have too many pixels (tuple)
    THRESHOLDS = (int(sys.argv[3]),97, 98, 99, 100) # tuple([ 25*pow(2,i) for i in range(5) ])
    # Remove single pixel spots. How many 8-connected pixels needed in a spot to keep it?
    PIXELS_IN_SPOT = 3
    # measurements that might be involved in scanning to transfer to output
    SCANMOTORS = ("diffrz","diffrz_center")
    # scalars from scan/instrument/positioners/XXXX to transfer to output
    HEADERMOTORS = ("diffty", "difftz" "samtx", "samty", "samtz", "bigy" )
    SCANTYPES = ("fscan", "f2scan", "fscan2d","finterlaced")
    #DETECTORNAME = "eiger"
    DETECTORNAME = "frelon3"
    MASKFILE = None # "/data/id11/nanoscope/Eiger/mask_20210428.edf"
    BACKGROUND = sys.argv[2]
    CORES = None  # guess the number of cores to use
    HNAME = sys.argv[1]
    OUTPATH = os.getcwd()

########################## should not need to change much below here
    def correct(self, frm):
        cor = frm
        if self.MASK is not None:
            cor = cor * self.MASK
        if self.BG is not None:
            cor = cor.astype(np.float32) - self.BG.estimate( cor )*bg_scale
        cor = np.multiply(cor,removalmask)
        print('Applied median subtraction and mask')
        return cor

# pip install ImageD11 --no-deps # if you do not have it yet:
from ImageD11 import sparseframe, cImageD11

class scaledbg(object):
    """
    data = scale * bg + offset
    frim ImageD11.scale (but with a pixel selection)
    """
    def __init__(self, bgimage, npx = 2048*8):
        order = np.argsort( bgimage.ravel() )
        self.bgimage = bgimage.copy().astype(np.float32)
        self.indx = np.concatenate( (order[:npx], order[-npx:]) )
        self.bgsample = self.bgimage.ravel()[ self.indx ]
        self.lsqmat = np.empty( (2,2), float )
        self.lsqmat[0,0] = np.dot(self.bgsample, self.bgsample)
        self.lsqmat[1,0] = self.lsqmat[0,1] = self.bgsample.sum()
        self.lsqmat[1,1] = len(self.bgsample)
        self.inverse = np.linalg.inv( self.lsqmat )
        self.bgestimage = np.empty_like( self.bgimage )

    def fit(self, other):
        imsample = other.ravel()[ self.indx ]
        rhs = ( np.dot( imsample, self.bgsample), imsample.sum() )
        return np.dot( self.inverse, rhs )

    def estimate(self, other):
        scale, offset = self.fit( other )
        # blas would be better
        np.multiply( self.bgimage, scale, self.bgestimage )
        np.add( self.bgestimage, offset, self.bgestimage )
        return self.bgestimage
    
def init(self):
    # validate input
    print("# max pixels",self.HOWMANY)
    self.CUT = self.THRESHOLDS[0]
    assert self.CUT >= 0
    for i in range(1, len(self.THRESHOLDS)):
        assert self.THRESHOLDS[i] > self.THRESHOLDS[i - 1]
    print("# thresholds", str(self.THRESHOLDS))
    if self.MASKFILE is not None:
        self.MASK = 1 - fabio.open(self.MASKFILE).data
        print("# Opened mask", self.MASKFILE)
    else:
        self.MASK = None
    if self.CORES is None:
        try:
            self.CORES = int(os.environ["SLURM_JOB_CPUS_PER_NODE"]) - 1
        except:
            self.CORES = os.cpu_count() - 1
    self.CORES = max(1, self.CORES)
    print("# Aiming to use", self.CORES, "processes")
    outh5 = os.path.split(self.HNAME)[-1].replace(".h5", f"_sparse.h5")
    self.OUTNAME = os.path.join(self.OUTPATH, outh5)
    print("# Output to ", self.OUTNAME)
    try:
        self.BG = scaledbg( fabio.open(self.BACKGROUND).data )
        print("# Background subtracted from",self.BACKGROUND)
    except:
        self.BG=None

Options.__init__ = init
# each subprocess gets their own options (no need if we fork?)
OPTIONS = Options()


# we will use multiple processes, each with 1 core
# all sub-processes better do this!
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"  # not sure we want/need this here?
os.environ["OMP_NUM_THREADS"] = "1"
os.environ["NUMBA_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"


@numba.njit
def select(img, mask, row, col, val, cut):
    # Choose the pixels that are > cut and put into sparse arrays
    k = 0
    for s in range(img.shape[0]):
        for f in range(img.shape[1]):
            if img[s, f] > cut:
                if mask[s, f]:  # skip masked
                    continue
                row[k] = s
                col[k] = f
                val[k] = img[s, f]
                k += 1
    return k

@numba.njit
def top_pixels(nnz, row, col, val, howmany, thresholds):
    """
    selects the strongest pixels from a sparse collection
    - THRESHOLDS should be a sorted array of potential cutoff values to try
    that are higher than the original cutoff used to select data
    - howmany is the maximum number of pixels to return
    """
    # quick return if there are already few enough pixels
    if nnz <= howmany:
        return nnz, thresholds[0]
    # histogram of how many pixels are above each threshold
    h = np.zeros(len(thresholds), dtype=np.uint32)
    for k in range(nnz):
        for i, t in enumerate(thresholds):
            if val[k] > t:
                h[i] += 1
            else:
                break
    # choose the one to use. This is the first that is lower than howmany
    tcut = thresholds[-1]
    for n, t in zip(h, thresholds):
        if n < howmany:
            tcut = t
            break
    # now we filter the pixels
    n = 0
    for k in range(nnz):
        if val[k] > tcut:
            row[n] = row[k]
            col[n] = col[k]
            val[n] = val[k]
            n += 1
            if n >= howmany:
                break
    return n, tcut

@functools.lru_cache(maxsize=1)
def get_dset(h5name, dsetname):
    """This avoids to re-read the dataset many times"""
    print(h5name)
    dset = h5py.File(h5name, "r")[dsetname]
    #dset = np.swapaxes(dset,1,2)
    return dset


def choose_parallel(args):
    """reads a frame and sends back a sparse frame"""
    h5name, address, frame_num = args
    print(frame_num)
    print(address)
    frm = get_dset(h5name, address)[frame_num]
    frm = np.swapaxes(frm,0,1) # This is wrong, seems already swapped when entering
    cor = OPTIONS.correct( frm )
    return frame_num, dosegment( cor )

def dosegment( frm ):
    if dosegment.cache is None:
        # cache the mallocs on this function. Should be one per process
        row = np.empty(frm.size, np.uint16)
        col = np.empty(frm.size, np.uint16)
        val = np.empty(frm.size, frm.dtype)
        if OPTIONS.MASK is None:
            msk = np.zeros( frm.shape, bool)
        else:
            msk = OPTIONS.MASK
        dosegment.cache = row, col, val, msk
    else:
        row, col, val, msk = dosegment.cache
    nnz = select(frm, msk, row, col, val, OPTIONS.CUT)
    if nnz == 0:
        sf = None
    else:
        if nnz > OPTIONS.HOWMANY:
            nnz, tcut = top_pixels(nnz, row, col, val, OPTIONS.HOWMANY,  OPTIONS.THRESHOLDS)
            #print("t=%d"%(tcut),end=' ')
        # Now get rid of the single pixel 'peaks'
        #   (for the mallocs, data is copied here)
        s = sparseframe.sparse_frame(row[:nnz].copy(), col[:nnz].copy(), frm.shape)
        s.set_pixels("intensity", val[:nnz].copy())
        if OPTIONS.PIXELS_IN_SPOT <= 1:
            sf = s
        else:
            # label them according to the connected objects
            sparseframe.sparse_connected_pixels(
                s, threshold=OPTIONS.CUT, data_name="intensity", label_name="cp"
            )
            # only keep spots with more than 3 pixels ...
            mom = sparseframe.sparse_moments(
                s, intensity_name="intensity", labels_name="cp"
            )
            npx = mom[:, cImageD11.s2D_1]
            pxcounts = npx[s.pixels["cp"] - 1]
            pxmsk = pxcounts >= OPTIONS.PIXELS_IN_SPOT
            if pxmsk.sum() == 0:
                sf = None
            else:
                sf = s.mask(pxmsk)
    return sf

dosegment.cache = None  # cache for malloc per process


def segment_scans( fname,
                   scans,
                   outname,
                   mypool,
                   scanmotors=OPTIONS.SCANMOTORS,
                   headermotors=OPTIONS.HEADERMOTORS,
                   detector=OPTIONS.DETECTORNAME ):
    """Does segmentation on a series of scans in hdf files:
    fname : input hdf file
    scans : input groups in the hdf files [h[s] for s in scans] will be treated
    outname : output hdf file to put the results into
    sparsify_func : function to select which pixels to keep
    headermotors : things to copy from input instrument/positioners to output
    scanmotors : datasets to copy from input measurement/xxx to output
    """
    opts = {
        "chunks": (10000,),
        "maxshape": (None,),
        "compression": "lzf",
        "shuffle": True,
    }
    ndone = 0
    with h5py.File(outname, "a") as hout:
        for scan in scans:
            if scan.endswith(".2"):  # for fscans
                continue
            with h5py.File(fname, "r") as hin:
                hout.attrs["h5input"] = fname
                gin = hin[scan]
                '''
                bad = 1
                if "title" not in hin[scan]:
                    print(scan,"missing title, skipping")
                    continue
                if "scan_data" not in hin[scan]:
                #if "measurement" not in hin[scan]:
                    print(scan,"missing measurement group, skipping")
                    continue
                if detector not in hin[scan]['measurement']:
                    print(scan,"missing measurement/%s, skipping"%(detector))
                    continue
                title = hin[scan]["title"][()]
                print("# ", scan, title)
                print("# time now", time.ctime(), "\n#", end=" ")
                for scantype in OPTIONS.SCANTYPES:
                    if title.find(scantype) >= 0:
                        bad = 0
                if ("measurement" in gin) and (detector not in gin["measurement"]):
                    bad += 1
                if bad:
                    print("# SKIPPING BAD SCAN", scan)
                    continue
                g = hout.create_group(scan)
                gm = g.create_group("measurement")
                for m in scanmotors:  # vary : many
                    if m in gin["measurement"]: #fix by add ["positioners"], Axel H. 24 feb 2022.
                        gm.create_dataset(m, data=gin["measurement"][m][:])
                        print("adding scanmotor: ", list(g["measurement"]))
                    else:
                        raise ValueError(str(m)+' is not in measurement'+ str( list(gin["measurement"]) ) )
                gip = g.create_group("instrument/positioners")
                for m in headermotors:  # fixed : scalar
                    if "instrument/positioners/%s"%(m) in gin:
                        gip.create_dataset(m, data=gin["instrument/positioners"][m][()])
                try:
                    frms = gin["measurement"][detector]
                except:
                    print(list(gin))
                    print(list(gin["measurement"]))
                    print(detector)
                    raise
                '''
                g = hout.create_group(scan)
                gm = g.create_group("measurement")
                tmp = np.array(range(3600))/10
                tmp2 = np.array([0.])
                gm.create_dataset("diffrz",data=tmp[:])
                gm.create_dataset("diffrz_center",data=tmp[:])
                print(np.array(g['measurement/diffrz']))
                gip = g.create_group("instrument/positioners")
                gip.create_dataset("diffty",data=tmp2[0])
                hout = h5py.File(fname,'r')
                print(hout.keys())
                print(fname)
                frms = hout['images']
                #frms = np.swapaxes(frms,1,2)
                #frms = hout['flyscan_00001/scan_data/images']
                row = g.create_dataset("row", (1,), dtype=np.uint16, **opts)
                col = g.create_dataset("col", (1,), dtype=np.uint16, **opts)
                # can go over 65535 frames in a scan
                num = g.create_dataset("frame", (1,), dtype=np.uint32, **opts)
                sig = g.create_dataset("intensity", (1,), dtype=frms.dtype, **opts)
                nnz = g.create_dataset("nnz", (frms.shape[0],), dtype=np.uint32)
                g.attrs["itype"] = np.dtype(np.uint16).name
                g.attrs["nframes"] = frms.shape[0]
                g.attrs["shape0"] = frms.shape[1]
                g.attrs["shape1"] = frms.shape[2]
                print("shape="+str(frms.shape[0]))
                print("gattr shape0="+str(g.attrs["shape0"]))
                npx = 0
                address = scan
                #address = scan + "/measurement/" + detector
                nimg = frms.shape[0]
                args = [(fname, address, i) for i in range(nimg)]
            # CLOSE the input file here
            # max chunksize is for load balancing. Originally tuned to be 7200/40/8 == 22
            max_chunksize = 64
            chunksize = min( max(1, len(args) // OPTIONS.CORES // 8), max_chunksize )
            # set the timeout to be 1 second per frame plus a minute. If we can't segment at
            # 1 f.p.s better to give up.
            timeout = len(args) + 60
            for i, spf in mypool.map(
                choose_parallel, args, chunksize=chunksize, timeout=timeout
            ):
                if i % 500 == 0:
                    if spf is None:
                        print("%4d 0" % (i), end=",")
                    else:
                        print("%4d %d" % (i, spf.nnz), end=",")
                    sys.stdout.flush()
                if spf is None:
                    nnz[i] = 0
                    continue
                if spf.nnz + npx > len(row):
                    row.resize(spf.nnz + npx, axis=0)
                    col.resize(spf.nnz + npx, axis=0)
                    sig.resize(spf.nnz + npx, axis=0)
                    num.resize(spf.nnz + npx, axis=0)
                row[npx:] = spf.row[:]
                col[npx:] = spf.col[:]
                sig[npx:] = spf.pixels["intensity"]
                num[npx:] = i
                nnz[i] = spf.nnz
                npx += spf.nnz
            ndone += nimg
            print("\n# Done", scan, nimg, ndone)
    return ndone

    # the output file should be flushed and closed when this returns


def main(hname, outname, mypool):
    # Remove old *sparse.h5 file if exit
    if os.path.exists(outname):
        os.remove(outname)
    # read all the scans : just the master process
    print("hname = ", repr(hname))
    print("outname = ", repr(outname))
    with h5py.File(hname, "r") as h:
        scans = list(h["/"])
        print("scans = ", repr(scans))
    print("#", time.ctime())
    start = time.time()
    print("outname= "+str(outname))
    nfrm = segment_scans(hname, scans, outname, mypool)
    end = time.time()
    print("#", time.ctime())
    print("# Elapsed", end - start, "/s,   f.p.s %.2f" % (nfrm / (end - start)))


if __name__ == "__main__":
    import concurrent.futures

    with concurrent.futures.ProcessPoolExecutor(max_workers=OPTIONS.CORES) as thepool:
        main(OPTIONS.HNAME, OPTIONS.OUTNAME, thepool)




import ImageD11.columnfile
import ImageD11.indexing
import scipy.spatial, scipy.sparse
import numpy as np, pylab as pl, pprint
import sys
import os

baseDir = "%s" % sys.argv[1]
str1 = "%s" % sys.argv[2]
str2 = "%s" % sys.argv[3]
str3 = "%s" % sys.argv[4]
merge_tol = float(sys.argv[5])

#str1 = "ottawa15_diff1/ottawa15_diff1_scan1/ottawa15_diff1_scan1.flt"
#str2 = "fitted_soleil_221029.par"
c = ImageD11.columnfile.columnfile(os.path.join(baseDir, str1))

#c = ImageD11.columnfile.columnfile(str1)
c.parameters.loadparameters(str2)

maxid = 96486
#pl.plot(c.omega,'.')
#pl.figure()
x=np.arange(20000,20150)
#x1 = range(maxid)
#pl.plot(x1,c.tth[0:maxid],'.')
#pl.plot(x1,c.tth[maxid:maxid*2],'+')
#pl.show()

# MUST MODIFY THIS FOR NEW DATA
maxid = int(round(float(c.omega.size/13))) # for diff1-scan1 TODO: fix this to make it automatic
c = c.copyrows(range(0,maxid))

# Sort by omega
c.sortby('omega')
omegavals = np.unique(c.omega)
omegavals.sort()

# try to merge the data over omega
n = [ (c.omega == o).sum() for o in omegavals]
p = np.cumsum( np.concatenate(([0,], n ) ) )
#pl.figure()
#pl.plot(n)
#pl.show()
# make a KDTree for each frame (wastes a bit of memory, but easier for sorting later)
trees = [ scipy.spatial.cKDTree( np.transpose( (c.sc[p[i]:p[i+1]], c.fc[p[i]:p[i+1]]) ) )
          for i in range(len(n)) ]
# peaks that overlap, k : 0 -> npks == len(s|f|o)
# diagonal
krow = list(range(c.nrows))
kcol = list(range(c.nrows))

for i in range(1,len(n)):  # match these to previous
    tlo = trees[i-1]
    thi = trees[i]
    # 1.6 is how close centers should be to overlap  Tian: 1.6--merge_tol
    lol = trees[i-1].query_ball_tree( trees[i], r = merge_tol )
    for srcpk, destpks in enumerate(lol):  # dest is strictly higher than src
        for destpk in destpks:
            krow.append( srcpk + p[i-1] )
            kcol.append( destpk + p[i] )
csr = scipy.sparse.csr_matrix( ( np.ones(len(krow), dtype=bool),
                               (kcol, krow) ), shape=(c.nrows,c.nrows) )
    # connected components == find all overlapping peaks
ncomp, labels= scipy.sparse.csgraph.connected_components( csr,
                                         directed=False, return_labels=True )
print(ncomp, labels.shape, c.nrows)

# now merge peaks across labels
c.titles
todo = [t for t in c.titles]
merged = {}
# things that are a simple sum
for item in 'Number_of_pixels', 'sum_intensity':
    data = c.getcolumn(item)
    merged[item] = np.bincount( labels, minlength=ncomp, weights=data)
# things that are an intensity weighted sum
ws = c.sum_intensity
wn = merged['sum_intensity']
for item in 'f_raw','s_raw','omega','sc','fc': #  'dty': # dty is not used so I am skipping it
    data = c.getcolumn(item) * ws
    wsum = np.bincount( labels, minlength=ncomp, weights=data)
    merged[item] = wsum / wn

mc = ImageD11.columnfile.colfile_from_dict( {t:merged[t] for t in c.titles if t in merged} )
mc.parameters.loadparameters(str2)
mc.updateGeometry()
mc.nrows, c.nrows
#pl.figure()
#pl.plot(c.omega, c.eta,'.')
#pl.plot(mc.omega, mc.eta,'+', ms=10)
#pl.xlim(0,10)
#pl.show()

# Plot things
#pl.plot(c.tth, c.eta,',')
#pl.plot(mc.tth, mc.eta,',')
#pl.show()

# Save things
mc.writefile(os.path.join(baseDir, str3))



import numpy as np
from hexrd import imageseries
from hexrd.imageseries.omega import OmegaWedges
import os
import h5py
import re
from tifffile import imread, TiffFile, imsave
import subprocess
import yaml
import sys
import argparse
import fabio
import time
import functools
import hdf5plugin
import numba
from ImageD11 import sparseframe, cImageD11
import pkg_resources

class HEDM_Toolkit:
    def __init__(self, **kwargs):
        self.params = kwargs
        self.input_file = self.params.get('input_file')
        self.imageseries_path = '/imageseries/images'
        self.image_default_path = '/images'
        self.image_input_path = self.params.get('image_input_path') 
        self.image_ilastik_path = '/exported_data'
        self.image_slice_path = None
        self.output_file = self.params.get('output_file')
        self.input_format = self._get_file_format()
        self.input_folder = self._get_input_folder()
        self.start_num = self.params.get('start_num', None)
        self.end_num = self.params.get('end_num', None) 
        self.num_images = self.params.get('num_images', None)
        self.omega = self.params.get('omega', None)
        self.bg_pct = self.params.get('bg_pct', None)
        self.bg_nf = self.params.get('bg_nf', None) 
        self.bgsub_h5 = self.params.get('bgsub_h5_file', None)
        self.bgsub = self.params.get('bgsub', True)
        self.empty_images = self.params.get('empty_images', 0)
        self.slice_file = self.params.get('slice_file', False)
        self.slice_images = self.params.get('slice_images', 0)
        self.ilastik_proc = self.params.get('ilastik_proc', True)
        self.ilastik_loc = self.params.get('ilastik_loc')
        self.ilastik_project_file = self.params.get('ilastik_project_file', None)
        self.generate_hexomap_files = self.params.get('generate_hexomap_files', None)
        self.tiff_path = self.params.get('tiff_path', None)
        self.prefix = self.params.get('prefix', None)
        self.tiff_output_folder = self.params.get('tiff_output_folder', None)
        self.hdf5_input_folder = self.params.get('hdf5_input_folder', None)
        self.input_tiff_folder = self.params.get('input_tiff_folder', None)
        self.selected_layers = self.params.get('selected_layers', None)
        self.input_hdf5 = self.params.get('input_hdf5', None)
        self.output_offset = self.params.get('output_offset', None)
        self.hdf5_file_name = self.params.get('hdf5_file_name', None)
        self.hdf5_file = self.params.get('hdf5_file', None)
        self.generate_hexrd_files = self.params.get('generate_hexrd_files', True)
        self.generate_ImageD11_files = self.params.get('generate_ImageD11_files', True)
        self.flip_option = self.params.get('flip_option', None)
        self.ImageD11_process = self.params.get('ImageD11_process', False)
        self.NPKS = self.params.get('NPKS', 50)
        self.UPKS = self.params.get('UPKS', 45)
        self.Tolseq1 = self.params.get('Tolseq1', 0.03)
        self.Tolseq2 = self.params.get('Tolseq2', 0.025)
        self.Toldist = self.params.get('Toldist', 200)
        self.flt_THRESHOLDS = self.params.get('flt_THRESHOLDS', 75)
        self.Idx_prefix = self.params.get('Idx_prefix', 'test0')
        self.input_file_ff = self.params.get('input_file_ff', None)
        self.input_filepath = self.params.get('input_filepath', None)
        self.dataset_path = self.params.get('input_filepath', None)
        self.symmetry = self.params.get('symmetry')
        self.ring1 = self.params.get('ring1')
        self.ring2 = self.params.get('ring2')
    

    def convert(self, *args, **kwargs):
        raise NotImplementedError("Subclass should implement this method")
    
    def _get_file_format(self):
        _, ext = os.path.splitext(self.input_file)
        return ext.lower()
    
    def _get_input_folder(self):
        return os.path.dirname(self.input_file)

    def tif2hdf5(self):
        input_file_prefix = os.path.splitext(os.path.basename(self.input_file))[0].rstrip("0123456789")
        image_paths = [os.path.join(self.input_folder, img_path) for img_path in os.listdir(self.input_folder) if (img_path.lower().endswith('.tif') or img_path.lower().endswith('.tiff')) and img_path.startswith(input_file_prefix)]

        if len(image_paths) == 0:
            print(f"No image paths found with the given prefix: {input_file_prefix}")
            return

        # Sort the image paths to ensure they are in the correct order
        image_paths = sorted(image_paths, key=lambda x: int(re.search(r'\d{6}', x).group()))

        if self.start_num is not None and self.end_num is not None:
            image_paths = image_paths[self.start_num:self.end_num+1]

        # Load the first image to get dimensions
        try:
            first_image = imread(image_paths[0])
        except Exception as e:
            print(f"Error loading the first image: {e}")
            return

        h, w = first_image.shape

        # Create an HDF5 file
        try:
            with h5py.File(self.output_file, 'w') as f:
                # Create a dataset with the shape (num_images, height, width) and the same dtype as the first image
                dataset = f.create_dataset('images', (len(image_paths), h, w), dtype=first_image.dtype)
                # Print the number of images in the dataset
                print(f"Number of images in the dataset: {len(image_paths)}")
                # Write the first image
                dataset[0] = first_image

                # Write the remaining images
                for i, img_path in enumerate(image_paths[1:], start=1):
                    try:
                        img = imread(img_path)
                    except Exception as e:
                        print(f"Error loading image {img_path}: {e}")
                        continue
                    dataset[i] = img
        except Exception as e:
            print(f"Error creating the HDF5 file: {e}")

    def make_meta():
        return {'testing': '1,2,3'}

    def ge2hdf5(self):
        # Define the width and height of the images.
        # The default is 2048*2048 for most beamlines.
        width = 2048
        height = 2048

        # Read the raw data from the GE file.
        with open(self.input_file, 'rb') as f:
            # Skip the header part of the file.
            # 8192 bytes is the header size for APS sector 1.
            f.read(8192)

            # Read the rest of the file as a numpy array.
            # The data type is np.uint16.
            data = np.fromfile(f, dtype=np.uint16)

        # Convert the read data to an array of images.
        try:
            image_data = np.reshape(data, (self.num_images, height, width))
        except ValueError:
            raise ValueError("Please input the right total number of frames: num_images")

        # Remove the specified number of empty images.
        if self.empty_images > 0:
            image_data = image_data[self.empty_images:]

        # Create an HDF5 file and write the image data to the file.
        with h5py.File(self.output_file, 'w') as f:
            # Create a dataset in the HDF5 file.
            # The shape of the dataset is (len(image_data), height, width).
            # The data type is the same as the image_data array.
            dataset = f.create_dataset('images', (len(image_data), height, width), dtype=image_data.dtype)
            
            # Print the number of images in the dataset.
            print(f"Number of images in the dataset: {len(image_data)}")
            
            # Write the images to the dataset.
            for i, img in enumerate(image_data):
                dataset[i] = img

    # May need to modified later for SOLEIL data
    # def hdf5_to_hdf5(self):
    #     # Copy the input HDF5 file to the output file with a new name and change the internal path to /images
    #     with h5py.File(self.input_file, 'r') as input_file:
    #         with h5py.File(self.output_file, 'w') as output_file:
    #             # Iterate over all keys in the input file
    #             for key in input_file.keys():
    #                 # Get the dataset from the input file
    #                 input_dataset = input_file[key]
    #                 # Remove the specified number of empty images
    #                 if self.empty_images > 0:
    #                     input_dataset = input_dataset[self.empty_images:]
    #                 # Create a new dataset in the output file with the same shape and dtype as the input dataset
    #                 # The new dataset is created at the '/images' path in the output file
    #                 output_dataset = output_file.create_dataset(f'/images', shape=input_dataset.shape, dtype=input_dataset.dtype)
    #                 # Copy the data from the input dataset to the output dataset
    #                 output_dataset[()] = input_dataset[()]

    def hdf5_to_hdf5(self):
        # Paths to check in the HDF5 file if dataset_path is not provided in the configuration
        possible_paths = ['/imageseries/images', '/images']

        # Open the input HDF5 file
        with h5py.File(self.input_file, 'r') as input_file:

            # If dataset_path is provided in the configuration, use it directly
            if self.dataset_path:
                if self.dataset_path not in input_file:
                    raise ValueError(f"Provided dataset_path {self.dataset_path} not found in {self.input_file}")
                dataset_path = self.dataset_path
            else:
                # Determine which path exists in the input_file
                dataset_path = None
                for path in possible_paths:
                    if path in input_file:
                        dataset_path = path
                        break

                # If no matching path is found, raise an error
                if not dataset_path:
                    raise ValueError(f"None of the expected paths {possible_paths} found in {self.input_file}")

            # Get the dataset from the determined path
            input_dataset = input_file[dataset_path]

            # Remove the specified number of empty images if needed
            if self.empty_images > 0:
                input_dataset = input_dataset[self.empty_images:]

            # Write to the output HDF5 file
            with h5py.File(self.output_file, 'w') as output_file:
                # Create a new dataset with the same shape and dtype as the input dataset
                # Store it under the '/images' path in the output file
                output_dataset = output_file.create_dataset('/images', shape=input_dataset.shape, dtype=input_dataset.dtype)
                # Copy the data from the input dataset to the output dataset
                output_dataset[()] = input_dataset[()]


    def bgsub4hdf5(self):
        ims = imageseries.open(
            self.output_file,
            format='hdf5',
            path='/',
            dataname='images'
        )

        # Input of omega meta data
        nf = self.num_images  #720
        omega = self.omega
        omw = OmegaWedges(nf)
        omw.addwedge(0, nf*omega, nf) 
        ims.metadata['omega'] = omw.omegas

        # Make dark image from first 100 frames
        pct = self.bg_pct
        nf_to_use = self.bg_nf
        dark = imageseries.stats.percentile(ims, pct, nf_to_use)
        # np.save(DarkFile, dark)

        # Now, apply the processing options
        ProcessedIS = imageseries.process.ProcessedImageSeries
        ops = [('dark', dark), ('flip', None)] # None, 'h', 'v', etc.
        pimgs = ProcessedIS(ims, ops)

        # Save the processed imageseries in hdf5 format
        imageseries.write(pimgs, self.bgsub_h5,'hdf5', path='/imageseries')

    def hdf5slice(self, input_file, slice_path):
        with h5py.File(input_file, 'r') as f:
            images = f[slice_path]
            base_name, extension = os.path.splitext(input_file)
            output_file = f"{base_name}_slice{extension}"
            with h5py.File(output_file, 'w') as f_example:
                example_images = f_example.create_dataset('images', (self.slice_images, images.shape[1], images.shape[2]), dtype=images.dtype)
                example_images[:] = images[:self.slice_images, :, :]

    @staticmethod 
    def get_tiff_compression(tiff_path):
        with TiffFile(tiff_path) as tif:
            return tif.pages[0].compression
         
    def sort_key(self):
        match = re.match(fr'{self.prefix}_layer(\d+)_det(\d+)_50bg_proc.h5', self.sort_filename)
        if not match:
            return float('inf')  # Place files that do not conform to the format at the end
        layer, det = map(int, match.groups())
        return layer * 2 + det  # Sort according to the numerical values of layer and det
    
    def get_start_num_from_filename(self):
        # Suppose your filename format is like 'prefix_layer1_det2_50bg_proc.h5'
        pattern = fr'nugget1_nf_layer(\d+)_det(\d+)_50bg_proc.h5'
        match = re.match(pattern, self.hdf5_file_name)
        if not match:
            raise ValueError(f"Invalid filename: {self.hdf5_file_name}")
        layer, det = map(int, match.groups())
        return (layer * 2 + det) * 180

    
    def hdf5_to_tiff(self):
        with h5py.File(self.hdf5_file, 'r') as f:
            dataset = f[self.image_ilastik_path]
            output_offset = self.get_start_num_from_filename()
            # The length of dataset
            num_images = len(dataset)
            # start and end idx
            start_idx = 0
            end_idx = num_images
            if start_idx >= num_images or end_idx > num_images:
                raise ValueError(f"Invalid range: start_num and end_num must be within the range of the dataset (0-{num_images - 1}).")
            input_tiff_folder = os.path.dirname(self.input_file)
            # get the tiff compression method for nf-hedm at aps
            first_tiff_path = [os.path.join(input_tiff_folder, p) for p in os.listdir(input_tiff_folder) if p.lower().endswith('.tif') or p.lower().endswith('.tiff')][0]
            compression = self.get_tiff_compression(first_tiff_path)
            start_idx = int(start_idx)
            end_idx = int(end_idx)
            for i in range(start_idx, end_idx):
                img = dataset[i].squeeze()
                img[img < 50] = 0
                img_filename = f"{self.prefix}_{i + output_offset:06d}.tif"
                img_path = os.path.join(tiff_output_folder, img_filename)
                imsave(img_path, img, compression=compression)

    def get_layer_from_filename(self):
        match = re.match(fr'{self.prefix}_layer(\d+)_det(\d+)_50bg_proc.h5', self.sort_filename)
        if not match:
            raise ValueError(f"Invalid filename: {self.sort_filename}")
        layer, det = map(int, match.groups())
        return layer
    
    def hdf5_to_npz(self, flip_option):
        base_name, extension = os.path.splitext(self.bgsub_h5)
        # Use the provided input_file_ff or the default if not provided
        if self.input_file_ff:
            input_file_ff = self.input_file_ff
            # Remove 'ilastik_' when input_file_ff is provided
            output_file = f"{base_name}_proc.npz"
            # If input_filepath is provided, strip the leading forward slash. Otherwise, use 'images' as the default.
            dataname = self.input_filepath.lstrip('/') if self.input_filepath else 'images'
        else:
            input_file_ff = f"{base_name}_ilastik_proc{extension}"
            output_file = f"{base_name}_ilastik_proc.npz"
            dataname = 'images'

        ims = imageseries.open(
            input_file_ff,
            format='hdf5',
            path='/',
            dataname=dataname
        )

        # Input of omega meta data
        nf = self.bg_nf  #720
        omega = self.omega
        omw = OmegaWedges(nf)
        omw.addwedge(0, nf*omega, nf) 
        ims.metadata['omega'] = omw.omegas

        # Make dark image from first 100 frames
        pct = self.bg_pct
        nf_to_use = self.bg_nf
        dark = imageseries.stats.percentile(ims, pct, nf_to_use)
        # np.save(DarkFile, dark)

        # Now, apply the processing options
        ProcessedIS = imageseries.process.ProcessedImageSeries
        ops = [('dark', dark), ('flip', flip_option)] # None, 'h', 'v', etc.
        pimgs = ProcessedIS(ims, ops)

        # Save the processed imageseries in npz format
        print(f"Writing npz file (may take a while): {output_file}")
        imageseries.write(pimgs, output_file, 'frame-cache', threshold=5, cache_file=output_file)

    ########################### 
    # ImageD11 related functions
    ###########################

    def hdf5_to_bg_edf(self):
        base_name, extension = os.path.splitext(self.bgsub_h5)
        
        # Use the provided input_file_ff or the default if not provided
        input_file_ff = self.input_file_ff if self.input_file_ff else f"{base_name}_ilastik_proc{extension}"
        
        # If input_filepath is not provided in the instance, use image_default_path
        if not self.input_filepath:
            input_filepath = self.image_default_path
        else:
            input_filepath = self.input_filepath
 
        # Open the ilastik processed file
        with h5py.File(input_file_ff, 'r') as f:
            im = f[input_filepath]
            im2 = np.median(im[:, :, :], axis=0)  # Calculate median for all frames
            
        # Save background
        output_filename = f"{base_name}_bg.edf"
        fabio.edfimage.edfimage(im2.astype(np.uint16)).write(output_filename)

    def run_sparse_script(self):
        script_path = pkg_resources.resource_filename('HEDM_Toolkit', 'scripts/sparse_h5.py')
        
        # Construct the input and output filenames
        base_name, extension = os.path.splitext(self.bgsub_h5)
        # Use the provided input_file_ff or the default if not provided
        input_file_ff = self.input_file_ff if self.input_file_ff else f"{base_name}_ilastik_proc{extension}"
        bg_filename = f"{base_name}_bg.edf"
        
        # Construct the full command
        cmd = [
            'python', script_path, 
            input_file_ff,
            bg_filename,
            str(self.flt_THRESHOLDS)
        ]
        
        # Execute the command
        subprocess.run(cmd)

    def run_gen_flt_script(self):
        script_path = pkg_resources.resource_filename('HEDM_Toolkit', 'scripts/gen_flt.py')
        dx_file = pkg_resources.resource_filename('HEDM_Toolkit', 'data/imaged11_dx.edf')
        dy_file = pkg_resources.resource_filename('HEDM_Toolkit', 'data/imaged11_dy.edf')
        
        # Construct the input and output filenames
        base_name, extension = os.path.splitext(self.bgsub_h5)
        # Use the provided input_file_ff or the default if not provided
        if self.input_file_ff:
            input_file_ff = self.input_file_ff
            # Remove 'ilastik_' when input_file_ff is provided
            sparse_file = f"{base_name}_proc_sparse.h5"
            flt_file = f"{base_name}_proc.flt"
        else:
            input_file_ff = f"{base_name}_ilastik_proc{extension}"
            sparse_file = f"{base_name}_ilastik_proc_sparse.h5"
            flt_file = f"{base_name}_ilastik_proc.flt"

        par_file = f"{base_name}.par"
        
        # Construct the full command
        cmd = [
            'python', script_path, 
            input_file_ff,
            sparse_file,
            flt_file,
            par_file,
            dx_file,  # Adding dx_file to the command
            dy_file   # Adding dy_file to the command
        ]
        # Execute the command
        subprocess.run(cmd)

    def run_clean_flt_script(self):
        script_path = pkg_resources.resource_filename('HEDM_Toolkit', 'scripts/clean_flt.py')

        # Construct the input and output filenames
        base_name, extension = os.path.splitext(self.bgsub_h5) 
        
        # Additional filenames based on your requirements
        merge3D_tor = 1.6
        par_file = f"{base_name}.par"
        base_dir = os.path.dirname(base_name)
        # Determine the flt_file_name based on the presence of input_file_ImageD11
        if self.input_file_ff:
            flt_file_name = os.path.basename(f"{base_name}_proc.flt")
        else:
            flt_file_name = os.path.basename(f"{base_name}_ilastik_proc.flt")        

        # Construct the full command
        cmd = [
            'python', script_path, 
            base_dir,
            flt_file_name,
            par_file,
            flt_file_name,  # Clean and overwrite .flt file
            str(merge3D_tor) # 1.6 as default, no need to modify
        ]
        
        # Execute the command
        subprocess.run(cmd)

    def run_ImageD11_indexing_script(self):
        script_path = pkg_resources.resource_filename('HEDM_Toolkit', 'scripts/indexing.py')

        # Construct the input and output filenames
        base_name, extension = os.path.splitext(self.bgsub_h5) 

        # Additional filenames based on your requirements
        par_file = f"{base_name}.par"
        Idx_file_prefix = self.Idx_prefix
        
        # Determine the flt_file_name based on the presence of input_file_ImageD11
        if self.input_file_ff:
            flt_file_name = os.path.basename(f"{base_name}_proc.flt")
        else:
            flt_file_name = os.path.basename(f"{base_name}_ilastik_proc.flt")

        NPKS = self.NPKS
        UPKS = self.UPKS
        Tolseq1 = self.Tolseq1
        Tolseq2 = self.Tolseq2
        Toldist = self.Toldist

        symmetry = self.symmetry
        ring1 = ",".join(map(str, self.ring1))
        ring2 = ",".join(map(str, self.ring2))

        # Set the OMP_NUM_THREADS environment variable
        os.environ['OMP_NUM_THREADS'] = '1'
        
        # Construct the full command
        cmd = [
            'python', script_path, 
            flt_file_name,
            par_file,
            Idx_file_prefix,
            str(NPKS),
            str(UPKS),
            str(Tolseq1),
            str(Tolseq2),
            str(Toldist),
            symmetry,
            ring1,
            ring2
        ]
        
        print("Executing command:", cmd)
        # Execute the command
        subprocess.run(cmd)

    def run_ImageD11_fitting_script(self, u_file, U_file, par_file, flt_file_name, t, omega_slop):
        script_path = pkg_resources.resource_filename('HEDM_Toolkit', 'scripts/fitting.py')
        cmd = [
            'python', script_path,
            '-u', u_file,
            '-U', U_file,
            '-p', par_file,
            '-f', flt_file_name,
            '-t', str(t),
            '--omega_slop', str(omega_slop)
        ]
        subprocess.run(cmd)

    def run_ImageD11_all_fitting_scripts(self):
        base_name, extension = os.path.splitext(self.bgsub_h5)
        par_file = f"{base_name}.par"
        # Determine the flt_file_name based on the presence of input_file_ff
        if self.input_file_ff:
            flt_file_name = os.path.basename(f"{base_name}_proc.flt")
        else:
            flt_file_name = os.path.basename(f"{base_name}_ilastik_proc.flt")
        t_values = [0.05, 0.04, 0.03, 0.02, 0.01]
        omega_slop = 0.05
        
        u_file = f'all{self.Idx_prefix}.map'
        for i, t in enumerate(t_values):
            U_file = f'all{self.Idx_prefix}_it{i+1}.map'
            self.run_ImageD11_fitting_script(u_file, U_file, par_file, flt_file_name, t, omega_slop)
            u_file = U_file

    def standard_hdf5(self, input_filepath, input_dataset_name, output_dataset_name):
        # Temporary output file path
        output_filepath = input_filepath + '_temp'
        # Open the original file
        with h5py.File(input_filepath, 'r') as f:
            data = f[input_dataset_name][:]  # load the data into memory
        # Remove the last dimension
        data_squeezed = np.squeeze(data, axis=-1)
        # Save the squeezed data to a new HDF5 file
        with h5py.File(output_filepath, 'w') as f:
            f.create_dataset(output_dataset_name, data=data_squeezed)
        # Delete the original file
        os.remove(input_filepath)
        # Rename the temporary output file to the original file name
        os.rename(output_filepath, input_filepath)

class Standardize_format(HEDM_Toolkit):
    def convert(self):
        print("Standardizing format...")
        if self.input_format in ('.ge', '.ge2', '.ge3', '.ge5'):
            print("Converting from .ge format to .h5...")
            self.ge2hdf5()
        elif self.input_format in ('.tif', '.tiff'):
            print("Converting from .tif format to .h5...")
            self.tif2hdf5()
        elif self.input_format in ('.hdf5', '.h5'):
            print("Input file is already in .h5 format. Copying to output file with a new name after deleting empty frames...")
            self.hdf5_to_hdf5()
        else:
            print(f"Unsupported input format: {self.input_format}")

class Subtract_background(HEDM_Toolkit):
    def convert(self):
        if self.bgsub:
            print("Applying background subtraction...")
            self.bgsub4hdf5()
        if self.slice_file:
            print("Generating sliced bgsub data...")
            self.hdf5slice(self.bgsub_h5, self.imageseries_path)

class Process_with_ilastik(HEDM_Toolkit):
    def convert(self):
        print(f'ilastik_project_file is {self.ilastik_project_file}')
        if self.ilastik_proc:
            print("Processing with ilastik...")
            base_name, extension = os.path.splitext(self.bgsub_h5) 
            output_file = f"{base_name}_ilastik_proc{extension}"
            
            if self.use_parallel_computing:     # If parallel computing is set to True in the config
                print("Running in parallel mode...")
                mpiexec_command = [
                    "mpirun", "-n", "12",
                    self.ilastik_loc,
                    "--headless",
                    "--distributed",
                    "--distributed-block-roi", '{"x": 2048, "y": 2048, "z": 1, "c": 2}',
                    f"--cutout_subregion=[(0,0,0,0),({self.bg_nf},2048,2048,1)]",
                    "--pipeline_result_drange=(0.0,1.0)",
                    "--export_drange=(0,100)",
                    f"--output_filename_format={output_file}",
                    f"--project={self.ilastik_project_file}",
                    "--export_source=Probabilities",
                    "--raw_data=" + f"{self.bgsub_h5}"
                ]
                subprocess.run(mpiexec_command)
            else:  # Otherwise, run the non-parallel code
                subprocess.run([
                    self.ilastik_loc,
                    '--headless',
                    f'--cutout_subregion=[(0,0,0,0),({self.bg_nf},2048,2048,1)]',
                    '--pipeline_result_drange=(0.0,1.0)',
                    '--export_drange=(0,100)',
                    f'--output_filename_format={output_file}',
                    f'--project={self.ilastik_project_file}',
                    '--export_source=Probabilities',
                    '--raw_data='+f'{self.bgsub_h5}'
                ], env={'LAZYFLOW_THREADS': '6'})
            
            self.standard_hdf5(output_file, self.image_ilastik_path, self.image_default_path)
    
        if self.slice_file:
            print("Generating sliced ilastik processed data...")
            self.hdf5slice(output_file, self.image_default_path)

class Convert_to_hedm_formats(HEDM_Toolkit):
    def convert(self):
        if self.generate_hexomap_files:
            print("Generating hexomap .tif files...")
            self.hdf5_to_tiff()

        if self.generate_hexrd_files:
            print("Generating hexrd .npz files...")
            self.hdf5_to_npz(self.flip_option)  
            print(f"HEXRD flip option is '{self.flip_option}'.")

        if self.generate_ImageD11_files:
            print("Preparing ImageD11 .edf bg files...")
            self.hdf5_to_bg_edf()
            print("Generating ImageD11 .h5 sparse files...")
            self.run_sparse_script()
            print("Generating ImageD11 .flt files...")
            self.run_gen_flt_script()
            print("Cleaning ImageD11 .flt files...")
            self.run_clean_flt_script()

        if not any([self.generate_hexomap_files, self.generate_hexrd_files, self.generate_ImageD11_files]):
            print("None of the generate flags is set to True. No action will be performed.")

class SliceHDF5(HEDM_Toolkit):
    def convert(self):
        print("Slicing HDF5 file...")
        self.hdf5slice(self.params.get('slice_input_file'), self.params.get('input_file_path'))

class ff_HEDM_process(HEDM_Toolkit):
    def convert(self):
        if self.ImageD11_process:
            print("ImageD11 indexing...")
            self.run_ImageD11_indexing_script()
            print("ImageD11 fitting...")
            self.run_ImageD11_all_fitting_scripts()

        elif self.hexrd_process:
            print("Hexrd find orientation...")
            self.hexrd_process()
            pass 
        elif self.x:
            print("x...")
            self.x()
            pass
        else:
            print("x.")


def run_function(converter_class, **params):
    converter = converter_class(**params)
    converter.convert()

def load_config(file_path):
    with open(file_path, 'r') as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return None


def main():
    parser = argparse.ArgumentParser(description='Process some parameters.')
    parser.add_argument('command', choices=['stand', 'sub', 'ilastik', 'hedm_formats', 'all', 'slice', 'ff_HEDM_process'])
    parser.add_argument('config_file')
    args = parser.parse_args()

    if len(sys.argv) != 3:
        print('Usage: HEDM_Toolkit <command> <config_file>')
        sys.exit()

    # Load configuration
    config_file = args.config_file
    params = load_config(config_file)
    if params is None:
        print('Error loading configuration')
        sys.exit()

    # Calculate bg_nf and add it to the params
    params['bg_nf'] = params['num_images'] - params['empty_images']

    # Create output directory only if generate_hexomap_files is True
    if params.get('generate_hexomap_files', False):
        os.makedirs(params['tiff_output_folder'], exist_ok=True)

    # Iterate through layers
    for layer in range(params['layers']):
        for det in range(params['dets']):
            params['output_file'] = params['base_dir'] + '{}_layer{}_det{}.h5'.format(params['sample_name'], layer, det)
            params['bgsub_h5_file'] = params['base_dir'] + '{}_layer{}_det{}_50bg.h5'.format(params['sample_name'], layer, det)

            start_num = params['start_constant'] + layer*params['num_images']*params['omega']*params['dets'] + det*params['num_images']*params['omega']
            end_num = start_num + params['num_images']*params['omega'] -1

            print("Currently executing for layer: {}, det: {}, start_num: {}".format(layer, det, start_num))  # print current layer, det and start_num

            params['start_num'] = start_num
            params['end_num'] = end_num

        # Call conversion function with the specific converter class you want to use
            if args.command == 'slice':
                run_function(SliceHDF5, **params)
            elif args.command == 'stand':
                run_function(Standardize_format, **params)
            elif args.command == 'sub':
                run_function(Subtract_background,  **params)
            elif args.command == 'ilastik':
                run_function(Process_with_ilastik, **params)
            elif args.command == 'hedm_formats':
                run_function(Convert_to_hedm_formats, **params)
            elif args.command == 'ff_HEDM_process':
                run_function(ff_HEDM_process, **params)
            elif args.command == 'all':
                run_function(Standardize_format, **params)
                run_function(Subtract_background,  **params)
                run_function(Process_with_ilastik, **params)
                run_function(Convert_to_hedm_formats, **params)
                run_function(ff_HEDM_process, **params)
if __name__ == "__main__":
    main()





import sys, os
# os.environ['OMP_NUM_THREADS']='1'  # use multiprocessing not openmp
import time
import h5py, numpy as np, numba, fabio, multiprocessing #, pylab as plt
from ImageD11 import unitcell
from ImageD11 import sparseframe, cImageD11, columnfile, transform
import fast_histogram
import matplotlib.pyplot as plt
from concurrent.futures import ProcessPoolExecutor
from ImageD11.sparseframe import sparse_frame
fnum  = 1 #10

def moments( bfrm ):
    I = bfrm[:, cImageD11.s2D_I]
    s = bfrm[:, cImageD11.s2D_sI] / I
    f = bfrm[:, cImageD11.s2D_fI] / I
    n = bfrm[:, cImageD11.s2D_1]
    a = I / n
    return s, f, a, n

def newpks(hf, scans=None, npixels=1, monitor=None, monval=1e3, ):
    """ do a peak search in a sparse frame series """
    titles = 's_raw f_raw avg_intensity Number_of_pixels sum_intensity omega dty'.split()
    c = {}
    for name in titles:
        c[name] = []
    # using the python file open overcomes some threading crap
    with h5py.File(open(hf,"rb"),'r') as hin:
        # scan numbers
        if scans is None:
            scans = list(hin['/'])
        for scan in scans:
            scan = 'images'
            gin = hin[scan]
            shape = gin.attrs['shape0'], gin.attrs['shape1']
            # import pdb; pdb.set_trace()
            omega = gin['measurement/diffrz'][:]
            difty = gin['instrument/positioners/diffty'][()]
            row = gin['row'][()]
            col = gin['col'][()]
            sig = gin['intensity'][()]
            if monitor is not None:
                mon = monval/gin[monitor][()]
            ipt = np.cumsum( gin['nnz'][:] )
            iprev = 0
            for k,nnz in enumerate( gin['nnz'][()] ) :
                inext = iprev + nnz
                if nnz == 0:
                    continue
                f = sparseframe.sparse_frame( row[iprev:inext],
                                              col[iprev:inext], shape )
                f.set_pixels("intensity", sig[iprev:inext] )
                sparseframe.sparse_connected_pixels(f, threshold=0.1)
                pks = sparseframe.sparse_moments(f, "intensity", "connectedpixels")
                #sparseframe.sparse_localmax(f)
                #pks = sparseframe.sparse_moments(f, "intensity", "localmax")
                s,f,a,n = moments( pks )
                m = n > npixels
                c['s_raw'].append( s[m] )
                c['f_raw'].append( f[m] )
                if monitor is not None:
                    c['avg_intensity'].append( a[m]*mon[k] )
                    c['sum_intensity'].append( a[m]*n[m]*mon[k] )
                else:
                    c['avg_intensity'].append( a[m] )
                    c['sum_intensity'].append( a[m]*n[m] )
                c['Number_of_pixels'].append( n[m] )
                npk = m.sum()
                c['omega' ].append( np.full( npk , omega[k] ) )
                c['dty'].append( np.full( npk , difty ) )
                iprev = inext
    for t in titles:
        c[t] = np.concatenate( c[t] )
    return c

def pfun( *args ):
#    print(args[0])
    s, h5name, npixels = args[0]
    return (s, newpks( h5name, [s,], npixels = npixels))

outname = "%s" % sys.argv[1]  #input h5 file
sparsename = "%s" % sys.argv[2]
filtername = "%s" % sys.argv[3]
par_file = "%s" % sys.argv[4]
dx_file = "%s" % sys.argv[5]
dy_file = "%s" % sys.argv[6] 

def sortedscans( hfo ):
    return [sc for _,sc in sorted([(float(s),s) for s in list(hfo)])]

#with h5py.File(outname,'r') as h:
    #im = h['flyscan_00001/scan_data/orca_image']
    #im3 = np.swapaxes(im,1,2)
    #bg_frames = np.median(im3[0::50,:,:],axis=0)
    #scans = sortedscans( h )
    #for scan in scans:
    #    r = h[scan]['measurement/diffrz'][()]
    #    print(scan,h[scan]['instrument/positioners/diffty'][()], r[0])
    #bg_frames = h[scan]['measurement/frelon3'][::100]

with h5py.File(outname, 'r') as hin:
    #frm = hin[scan]['measurement/frelon3'][fnum]
    im = hin['images']
    #im3 = np.swapaxes(im,1,2)
    frm = im[fnum,:,:]

with h5py.File(sparsename,"r") as hin:
    scan='images'
    print(list(hin[scan]), dict(hin[scan].attrs))
    print(list(hin[scan]['instrument']))
    f = fnum
    nnz = hin[scan]['nnz'][:]
    ipt = np.cumsum(nnz)
    # num_frames = len(hin[scan]['frame'])

    # if num_frames == 2:
    #     # Adjust the logic for 2 frames
    #     # For example, if you know `f` should be 1 for the second frame, set it manually:
    #     f = 1
    #     print(f, ipt[f-1], hin[scan]['frame'][ipt[f-1]])
    # else:
    #     # The original logic for more frames
    #     print(f, ipt[f-1], ipt[f-2], hin[scan]['frame'][ipt[f-1]], hin[scan]['frame'][ipt[f]])

    s = ipt[f-1]
    e = ipt[f]
    g = hin[scan]
    sh = g.attrs['shape0'],g.attrs['shape1']
    spf = sparse_frame( g['row'][s:e], g['col'][s:e], shape = (g.attrs['shape0'],g.attrs['shape1']),
                       pixels={'intensity': g['intensity'][s:e]})
    seg = spf.to_dense('intensity')

class spatial(object):
    def __init__(self,
                 dxfile=dx_file,
                 dyfile=dy_file,):
        import fabio
        self.dx = fabio.open(dxfile).data
        self.dy = fabio.open(dyfile).data
    def __call__(self, pks):
        si = np.round(pks['s_raw']).astype(int)
        fi = np.round(pks['f_raw']).astype(int)
        pks['sc'] = self.dy[ si, fi ] + pks['s_raw']
        pks['fc'] = self.dx[ si, fi ] + pks['f_raw']
        pks['sc'] = pks['s_raw']
        pks['fc'] = pks['f_raw']
        return pks

def tocolf(pks):
    titles = list(pks.keys())
    colf = columnfile.newcolumnfile( titles=titles )
    nrows = len(pks[titles[0]])
    colf.nrows = nrows
    colf.set_bigarray( [ pks[t] for t in titles ] )
    return colf

spat = spatial(  )

sparsename_load = sparsename

# Open hdf5 file of sparse peaks
with h5py.File(sparsename_load,'r') as h:
    #scans_load = sortedscans( h )
    #for scan in scans_load:
    scans = 'images'
    scans_load = 'images'
    #r = h[scan]['measurement/diffrz'][()]
    #print(scan,h[scan]['instrument/positioners/diffty'][()], r[0])

print('scans_load='+scans_load)
pks_load = newpks(sparsename_load, scans=scans_load, npixels=2, monitor=None)

# Save flt filtered peaks file
pks_load.keys(), 12.3985/42.5
pks_load = spat(pks_load)
colf_load = tocolf(pks_load)
colf_load.filter(colf_load.Number_of_pixels>5)
colf_load.parameters.loadparameters(par_file)
colf_load.updateGeometry()
colf_load.filter(colf_load.Number_of_pixels*np.exp(0.4*colf_load.ds**2)>40)
#plt.figure()
#plt.plot(colf_load_02.tth, colf_load_02.eta,'.',alpha=0.3)

# Get number of lines in filtered peaks file
print('num spots = '+str(colf_load.nrows))
uc_load = unitcell.unitcell_from_parameters(colf_load.parameters)
uc_load.makerings(colf_load.ds.max())
s2 = colf_load.ds
colf_load.filter( colf_load.ds < 0.75)
la = colf_load.copy()
la.filter( colf_load.ds < 0.75)
la.writefile(filtername)
#m = abs(colf_load.ds - 0.23) < 0.01
#plt.figure()
#plt.plot(180-colf_load.eta, 180+colf_load.omega, 'o', alpha=0.2)

# Set the minimal things to save
minimal = { k:colf_load[k] for k in ['sc','fc','omega','dty','sum_intensity','Number_of_pixels']}
colf_load = minimal

# Create random number
def rnd( a, p, t):
    if p == 0:
        return a.astype(t)
    # hack to try to get floats to compress better. Round to nearest integer p
    a = (np.round(a*pow(2,p)).astype(np.int).astype(t))/pow(2,p)
    return a
# # Save sparse peaks hdf5 file
# with h5py.File(pksname,'w') as hout:
#     g = hout.create_group('peaks')
#     g.attrs['ImageD11_type'] = 'peaks'
#     for name,b,t in ( ('sc',8, np.float32),
#                  ('fc',8, np.float32),
#                  ('omega',8, np.float32),
#                  ('dty',8, np.float32),
#                  ('sum_intensity',2, np.float32),
#                  ('Number_of_pixels',0, np.uint32)):
#         a = rnd(colf_load[name],b,t)
#         print(name, 'chunks', a.size/pow(2,11))
#         g.create_dataset( name,
#                           data = a,
#                           dtype = t,
#                           compression = 'lzf',
#                           shuffle = True,
#                           chunks = ( pow(2,11) ,), #  32 kB is
#                         )



#!/usr/bin/python3

from __future__ import print_function

from ImageD11.indexing import readubis, write_ubi_file
from ImageD11.refinegrains import refinegrains
import ImageD11.refinegrains
from ImageD11 import ImageD11options
import sys, os, argparse


def makemap(options):
    try:
        if options.tthrange is None:
            tthr = (0.,180.)
            #tthr = (9.5,11.85)
        else:
            tthr = options.tthrange
            if len(tthr) == 1:
                tthr = ( 0, tthr[0] )
            print("Using tthrange",tthr)
        func = getattr(ImageD11.refinegrains, options.latticesymmetry )
        o = refinegrains(intensity_tth_range = tthr,
                         latticesymmetry = func,
                         OmFloat=options.omega_float,
                         OmSlop=options.omega_slop)
    except:
        raise
        o = refinegrains()
    o.loadparameters(options.parfile)
    print("got pars")
    o.loadfiltered(options.fltfile)
    print("got filtered")
    o.readubis(options.ubifile)
    if options.symmetry != "triclinic":
        # Grainspotter will have already done this
        print("transform to uniq")
        o.makeuniq(options.symmetry)
    print("got ubis")
    o.tolerance = float(options.tol)
    print("generating")
    print(o.parameterobj.parameters)
    #o.parameterobj.parameters['t_x']=2000
    o.generate_grains()
    #print(o.parameterobj.parameters)
    #o.parameterobj.parameters['t_x']=2000
    #o.parameterobj.parameters['t_x']=2000
    print("Refining posi too")
    # o.refineubis(quiet = False , scoreonly = True)
    print("Refining positions too")
    o.refinepositions()
    print("Done refining positions too")
    # o.refineubis(quiet = False , scoreonly = True)
    o.savegrains(options.newubifile, sort_npks = options.sort_npks)
    col = o.scandata[options.fltfile].writefile( options.fltfile+".new")
    if hasattr(options, "newfltfile") and options.newfltfile is not None:
        print("re-assignlabels")
        o.assignlabels()
        col = o.scandata[options.fltfile].copy()
        print("Before filtering",col.nrows)
        col.filter(col.labels < -0.5)
        # print col.labels[:10]
        print("After filtering",col.nrows)
        col.writefile(options.newfltfile)


def get_options(parser):
    parser = ImageD11.refinegrains.get_options(parser)
    parser.add_argument("--no_sort", action="store_false",
                      dest="sort_npks", default = True,
                      help="Sort grains by number of peaks indexed")
    parser.add_argument( "--tthrange", action="append",
                      dest = "tthrange", type=float,
                      default = None,
                      help= "Two theta range for getting median intensity")
    return parser


if __name__ == "__main__":
    import logging, sys

    console = logging.StreamHandler(sys.stdout)
    # set a format which is simpler for console use
    formatter = logging.Formatter('%(levelname)-8s : %(message)s')
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    console.setLevel(logging.DEBUG)
    root = logging.getLogger('')
    root.addHandler(console)
    root.setLevel(logging.DEBUG) # should we process everything...?


    parser = get_options( argparse.ArgumentParser() )

    options = parser.parse_args()

    for name in ["parfile" ,
                 "ubifile",
                 "newubifile",
                 "fltfile"]:
        if getattr(options, name) is None:
            parser.print_help()
            logging.error("Missing option "+name)
            import sys
            sys.exit()

    try:
        makemap(options)
    except:
        parser.print_help()
        import traceback
        logging.error("An error occurred, details follow")
        traceback.print_exc()
 


from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='HEDM_Toolkit',
    version='0.1.2',
    author='Ye Tian',
    author_email='ytian37@jhu.edu',
    entry_points={
        'console_scripts': [
            'HEDM_Toolkit=HEDM_Toolkit.HEDM_Toolkit:main',
            'copy_demo=HEDM_Toolkit.utilities:copy_demo_func', # assuming the function is in utilities module
        ],
    },
    description='Tools for High Energy Diffraction Microscopy (HEDM) analysis',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/minety/HEDM',
    packages=find_packages(),
    package_data={
        'HEDM_Toolkit': ['scripts/*', 'data/*', 'Demo/*'],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    python_requires='>=3.8'
)
